#ifndef _FLUIDS_H_
#define _FLUIDS_H_


//////////////////////////////////////////////////////////////////////////////
// Includes

#include "simplicialComplex/simplicialComplex.h"
#include "fluids/dec.h"
#include "fluids/polyhedron.h"
#include "fluids/meshPoint.h"


//////////////////////////////////////////////////////////////////////////////

class Tet;


//////////////////////////////////////////////////////////////////////////////
// class Fluid
// Abstract base class for the 2D and 3D versions

class Fluid
{
public:  // methods
    Fluid();
    virtual ~Fluid();

    virtual bool LoadMesh( const char *filename, bool freeBoundaries = false ) = 0;

    void StepSimulation( void );

    void SetTimestep               ( real dt )  { m_timestep = dt; }
    void SetNumParticles           ( int n )    { m_numParticles = n; }
    void SetNumStreamlines         ( int n )    { m_numStreamlines = n; }
    void SetNumPointsPerStreamline ( int n )    { m_numPointsPerStreamline = n; }
    void SetDiffusionCoefficient   ( real mu )  { m_diffusionCoefficient = mu; }
    void SetParticleBouyancy       ( real b )   { m_particleBouyancy = b; }

    void DumpParticlesToFile( const char *filename );

protected:  // methods
    void InitializeVizArrays( void );

    void BeginStreamline    ( void );
    void AddStreamlinePoint ( const vec3& point );
    void EndStreamline      ( void );

    virtual void AdvectVorticities  ( void ) = 0;
    virtual void AdvectParticles    ( void ) = 0;
    virtual void AddForces          ( void ) = 0;
    virtual void FluxesToVorticities( void ) = 0;
    virtual void VorticitiesToFluxes( void ) = 0;
    virtual void FluxesToVelocities ( void ) = 0;

protected:  // data
    real   m_timestep;
    real   m_diffusionCoefficient;
    real   m_particleBouyancy;

    size_t m_numParticles;

    size_t m_numStreamlines;
    size_t m_numPointsPerStreamline;
    size_t m_currentStreamline;
    size_t m_curSlinePoint;

    std::vector< vec3 >  m_velocityField;
    std::vector< vec3 >  m_velocityFieldVertices;

public:  // Viz stuff
    std::vector< vec3f >                 m_particles;
    std::vector< std::vector< vec3f > >  m_streamlines;
    std::vector< std::vector< vec3f > >  m_slineTangents;
    std::vector< vec3f >                 m_colors;
};


//////////////////////////////////////////////////////////////////////////////
// class Fluid3D

class Fluid3D : public Fluid
{
    friend class FlowViz;

public:  // methods
    Fluid3D();
    ~Fluid3D();

    bool LoadMesh( const char *filename, bool freeBoundaries = false );

    const SimplicialComplex<3>& Mesh( void ) { return m_mesh; }


private:  // methods
    Fluid3D( const Fluid3D & );
    Fluid3D &operator=( const Fluid3D & );

    void InitializeArrays    ( void );
    void PlaceParticles      ( void );
    void BuildDualCells      ( void );
    void BuildTetToDualMap   ( void );
    void BuildWhitneyBasis   ( void );
    void BuildOrLoadMatrices ( bool freeBoundaries, const char *meshFilename );

    void AdvectParticles     ( void );
    //void AdvectDensities     ( void );
    void AdvectVorticities   ( void );
    void AddForces           ( void );

    vec3 GetCirculation      ( const MeshPoint3& point );

    void PathTrace           ( bool negativeDirection, MeshPoint3& point, bool updateStreamlines = false );
    void PathTraceSmooth     ( bool negativeDirection, MeshPoint3& point, bool updateStreamlines = false );
    real TraceTet            ( const Tet& tet, const vec4& posBary, const vec4& dirBary, Tet& nextTet );

    void FluxesToVorticities ( void );
    void VorticitiesToFluxes ( void );
    void FluxesToVelocities  ( void );


private:  // data
    SimplicialComplex<3> m_mesh;

    Form<PRIMAL,2>  m_fluxes;
    Form<PRIMAL,1>  m_potentials;
    Form<DUAL,2>    m_vorticities;
    Form<DUAL,2>    m_forces;
    Form<DUAL,2>    m_scratchD2;

    //std::vector<real>  m_densities;

    std::vector< vec4 >         m_velocityFieldBary;

    std::vector< vec<4,vec3> >  m_whitneyBasis;
    std::vector< vec4 >         m_fluxesToBaryVecs;

    std::vector<Polyhedron>     m_dualCells;
    std::vector<MeshPoint3>     m_dualVertices;
    std::vector<MeshPoint3>     m_particleTracker;

    std::vector< std::vector<size_t> > m_tetToDualCellMap;

    std::vector<real>           m_baryWeights;

    DecOp<PRIMAL,1,PRIMAL,2>    d_P1;
    DecOp<PRIMAL,2,DUAL,2>      d_star_P2;
    DecOp<PRIMAL,1,DUAL,2>      sdsdspdsd_P1;  // star_d_star_d_star_plus_d_star_d_P1;
};



#endif  // _FLUIDS_H_
