//////////////////////////////////////////////////////////////////////////////
// LinearAlgebra.cpp
// Defines the Vector class and the Sparse Matrix class.
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Includes

#include "linearAlgebra/linearAlgebra.h"
#include "linearAlgebra/linAlgImpl.h"


//////////////////////////////////////////////////////////////////////////////

// Change this line to turn on/off Petsc use
#define USE_PETSC 1

// (but if you didn't compile with it, you have no choice)
#ifndef COMPILE_WITH_PETSC
    #undef  USE_PETSC
    #define USE_PETSC 0
#endif


//////////////////////////////////////////////////////////////////////////////
// Vector public functions
// Passes everything on to impl

Vector::Vector( size_t size )
{
#if USE_PETSC
    m_impl = new PetscVectorImpl( size );
#else
    m_impl = new MyVectorImpl( size );
#endif
}
Vector::~Vector()
{
    delete m_impl;
}
double & Vector::operator [] ( size_t i )
{
    return m_impl->operator []( i );
}
const double & Vector::operator [] ( size_t i ) const
{
    return m_impl->operator []( i );
}
void Vector::Assign( size_t size, double value )
{
    m_impl->Assign( size, value );
}
void Vector::Set( double value )
{
    m_impl->Set( value );
}
void Vector::Copy( const Vector &v )
{
    m_impl->Copy( v.m_impl );
}
void Vector::Swap( Vector &v )
{
    m_impl->Swap( v.m_impl );
}
void Vector::Add( const Vector &v )
{
    m_impl->Add( v.m_impl );
}
void Vector::Subtract( const Vector &v )
{
    m_impl->Subtract( v.m_impl );
}
void Vector::Multiply( const Vector &v )
{
    m_impl->Multiply( v.m_impl );
}
void Vector::Multiply( real f )
{
    m_impl->Multiply( f );
}
void Vector::Divide( const Vector &v )
{
    m_impl->Divide( v.m_impl );
}
void Vector::Negate()
{
    m_impl->Negate();
}
void Vector::MAD( real f, const Vector &v )
{
    m_impl->MAD( f, v.m_impl );
}
void Vector::MAD( const Vector &v, real f )
{
    m_impl->MAD( v.m_impl, f );
}
real Vector::DotProduct( const Vector &v ) const
{
    return m_impl->DotProduct( v.m_impl );
}


//////////////////////////////////////////////////////////////////////////////
// SparseMatrix public functions
// Passes everything on to impl

SparseMatrix::SparseMatrix()
{
#if USE_PETSC
    m_impl = new PetscMatImpl;
#else
    m_impl = new MyMatImpl;
#endif
}
SparseMatrix::SparseMatrix( int numRows, int numColumns )
{
#if USE_PETSC
    m_impl = new PetscMatImpl( numRows, numColumns );
#else
    m_impl = new MyMatImpl( numRows, numColumns );
#endif
}
SparseMatrix::~SparseMatrix()
{
    delete m_impl;
}
void SparseMatrix::BeginConstruction( int numRows, int numColumns, bool bSymmetric )
{
    m_impl->BeginConstruction( numRows, numColumns, bSymmetric );
}
void SparseMatrix::EndConstruction( PRECOND_TYPE_ENUM type, real dropTolerance )
{
    m_impl->EndConstruction( type, dropTolerance );
}
void SparseMatrix::Save( const char *filename )
{
#if USE_PETSC
    m_impl->Save( (std::string("petsc_") + filename).c_str() );
#else
    m_impl->Save( filename );
#endif
}
void SparseMatrix::Load( const char *filename )
{
#if USE_PETSC
    m_impl->Load( (std::string("petsc_") + filename).c_str() );
#else
    m_impl->Load( filename );
#endif
}
void SparseMatrix::Set( int i, int j, real value )
{
    m_impl->Set( i, j, value );
}
void SparseMatrix::Add( int i, int j, real value )
{
    m_impl->Add( i, j, value );
}
void SparseMatrix::Multiply( const Vector &vec, Vector &product ) const
{
    m_impl->Multiply( vec.m_impl, product.m_impl );
}
bool SparseMatrix::Solve( const Vector &b, Vector &x, int maxIterations, real epsilon ) const
{
    return m_impl->Solve( b.m_impl, x.m_impl, maxIterations, epsilon );
}
void SparseMatrix::LogConvergenceStats( bool enable )
{
    m_impl->LogConvergenceStats( enable );
}
