
//////////////////////////////////////////////////////////////////////////////
// Includes

#include "simplicialComplex/tet.h"
#include "simplicialComplex/simplicialComplex.h"
#include "simplicialComplex/vertex.h"
#include "simplicialComplex/face.h"
#include "util/assert.h"


//////////////////////////////////////////////////////////////////////////////
// Default Constructor

Tet::Tet() : sc(0), index(size_t(-1))
{
    // empty
}

//////////////////////////////////////////////////////////////////////////////
// Private Constructor that takes a mesh pointer and index

Tet::Tet( const SimplicialComplex<3> *sc, size_t index ) : sc(sc), index(index)
{
    // empty
}

//////////////////////////////////////////////////////////////////////////////
// Destructor

Tet::~Tet()
{
    // empty
}

//////////////////////////////////////////////////////////////////////////////
// operator bool - Is this face valid

Tet::operator bool() const
{
    return (sc && index < sc->NumTets());
}

//////////////////////////////////////////////////////////////////////////////
// Center

vec3 Tet::Center() const
{
    smart_assert(*this).msg( "Invalid tet" ).fatal();
    return sc->m_tets[index].Center();
}

//////////////////////////////////////////////////////////////////////////////
// Volume

real Tet::Volume() const
{
    smart_assert(*this).msg( "Invalid tet" ).fatal();
    return sc->m_tets[index].Volume();
}

//////////////////////////////////////////////////////////////////////////////
// Vertices

void Tet::Vertices( vec3& v0, vec3& v1, vec3& v2, vec3& v3 ) const
{
    smart_assert(*this).msg( "Invalid tet" ).fatal();
    v0 = sc->m_vertices[ sc->m_tets[index].Get(0) ].Center();
    v1 = sc->m_vertices[ sc->m_tets[index].Get(1) ].Center();
    v2 = sc->m_vertices[ sc->m_tets[index].Get(2) ].Center();
    v3 = sc->m_vertices[ sc->m_tets[index].Get(3) ].Center();
}

//////////////////////////////////////////////////////////////////////////////
// Vertices

void Tet::Vertices( int& v0, int& v1, int& v2, int& v3 ) const
{
    smart_assert(*this).msg( "Invalid tet" ).fatal();
    v0 = sc->m_tets[index].Get(0);
    v1 = sc->m_tets[index].Get(1);
    v2 = sc->m_tets[index].Get(2);
    v3 = sc->m_tets[index].Get(3);
}

//////////////////////////////////////////////////////////////////////////////
// Vertices

void Tet::Vertices( Vertex& v0, Vertex& v1, Vertex& v2, Vertex& v3 ) const
{
    smart_assert(*this).msg( "Invalid tet" ).fatal();
    v0.sc = v1.sc = v2.sc = v3.sc = sc;
    v0.index = sc->m_tets[index].Get(0);
    v1.index = sc->m_tets[index].Get(1);
    v2.index = sc->m_tets[index].Get(2);
    v3.index = sc->m_tets[index].Get(3);
}

//////////////////////////////////////////////////////////////////////////////
// Faces

void Tet::Faces( Face& f0, Face& f1, Face& f2, Face& f3 ) const
{
    smart_assert(*this).msg( "Invalid tet" ).fatal();
    f0.sc = f1.sc = f2.sc = f3.sc = sc;
    f0.index = sc->m_TFmatrix[ index ][ 0 ].first;
    f1.index = sc->m_TFmatrix[ index ][ 1 ].first;
    f2.index = sc->m_TFmatrix[ index ][ 2 ].first;
    f3.index = sc->m_TFmatrix[ index ][ 3 ].first;
}

//////////////////////////////////////////////////////////////////////////////
// OppositeVertex

Vertex Tet::OppositeVertex( const Face& face ) const
{
    smart_assert(*this).msg( "Invalid tet" ).fatal();
    smart_assert(face).msg( "Invalid face" ).fatal();

    Vertex v;
    v.sc = sc;
    if( face.Index() == sc->m_TFmatrix[ index ][ 0 ].first )
        v.index = sc->m_tets[index].Get(0);
    else if( face.Index() == sc->m_TFmatrix[ index ][ 1 ].first )
        v.index = sc->m_tets[index].Get(1);
    else if( face.Index() == sc->m_TFmatrix[ index ][ 2 ].first )
        v.index = sc->m_tets[index].Get(2);
    else if( face.Index() == sc->m_TFmatrix[ index ][ 3 ].first )
        v.index = sc->m_tets[index].Get(3);
    else
        v.index = size_t(-1);

    return v;
}

//////////////////////////////////////////////////////////////////////////////
// AdjacentTet - across from face

Tet Tet::AdjacentTet( const Face& face ) const
{
    smart_assert(*this).msg( "Invalid tet" ).fatal();
    smart_assert(face).msg( "Invalid face" ).fatal();

    Tet t0, t1;
    face.Tets( t0, t1 );
    if( index == t0.index )
        return t1;
    if( index == t1.index )
        return t0;
    return Tet();
}

//////////////////////////////////////////////////////////////////////////////
// AdjacentTet - given a neighbor number, 0-3

Tet Tet::AdjacentTet( int whichNeighbor ) const
{
    smart_assert(*this).msg( "Invalid tet" ).fatal();
    smart_assert(whichNeighbor >=0 && whichNeighbor < 4)(whichNeighbor).msg( "Invalid neighbor index" ).fatal();

    Face face( sc, sc->m_TFmatrix[ index ][ whichNeighbor ].first );
    Tet t0, t1;
    face.Tets( t0, t1 );
    if( index == t0.index )
        return t1;
    if( index == t1.index )
        return t0;
    return Tet();
}

//////////////////////////////////////////////////////////////////////////////
// IsOnBoundary

bool Tet::IsOnBoundary() const
{
    smart_assert(*this).msg( "Invalid tet" ).fatal();
    Face f0, f1, f2, f3;
    Faces( f0, f1, f2, f3 );
    return ( f0.IsOnBoundary() || f1.IsOnBoundary() || f2.IsOnBoundary() || f3.IsOnBoundary() );
}

//////////////////////////////////////////////////////////////////////////////
// Equality operator

bool operator==( const Tet& lhs, const Tet& rhs )
{
    return (lhs.sc == rhs.sc && lhs.index == rhs.index);
}

//////////////////////////////////////////////////////////////////////////////
// Inequality operator

bool operator!=( const Tet& lhs, const Tet& rhs )
{
    return (!(lhs == rhs));
}
