#ifndef _FUNCTION_H_
#define _FUNCTION_H_

#include "util/assert.h"


/**
* @struct function
* @ingroup util
*
* This is a function object wrapper. It can refer to either a function pointer
* or a functor. Supports signatures with 0, 1, or 2 parameters.
*/
template< typename R, typename T1 = void, typename T2 = void > struct function;


//////////////////////////////////////////////////////////////////////////////
// 2 parameters

template< typename R, typename T1, typename T2 >
struct functor_caller_base
{
    virtual ~functor_caller_base() {}
    virtual R operator()( T1, T2 ) const = 0;
    virtual functor_caller_base* clone() const = 0; 
};

template< typename F, typename R, typename T1, typename T2 >
struct functor_caller : functor_caller_base<R,T1,T2>
{
    functor_caller( F f ) : f(f) {}
    virtual R operator()( T1 p1, T2 p2 ) const { return f( p1, p2 ); }
    virtual functor_caller_base<R,T1,T2>* clone() const { return new functor_caller<F,R,T1,T2>(f); } 
private:
    F f;
};

template< typename R, typename T1, typename T2 >
struct function
{
    function() : fc(0) {}
    ~function() { delete fc; }
    template< typename F > function( F f ) { fc = new functor_caller<F,R,T1,T2>(f); }
    function( const function& r ) : fc(0) { operator=(r); }
    function& operator=( const function& r )
    {
        if( this == &r ) return *this;
        delete fc;
        fc = r.fc->clone();
        return *this;
    }
    R operator()( T1 p1, T2 p2 ) const
    {
        smart_assert(fc).msg( "Uninitialized function object" ).fatal();
        return (*fc)( p1, p2 );
    }
private:
    functor_caller_base<R,T1,T2> *fc;
};


//////////////////////////////////////////////////////////////////////////////
// 1 parameter

template< typename R, typename T1 >
struct functor_caller_base< R, T1, void >
{
    virtual ~functor_caller_base() {}
    virtual R operator()( T1 ) const = 0;
    virtual functor_caller_base* clone() const = 0; 
};

template< typename F, typename R, typename T1 >
struct functor_caller<F,R,T1,void> : functor_caller_base<R,T1,void>
{
    functor_caller( F f ) : f(f) {}
    virtual R operator()( T1 p1 ) const { return f( p1 ); }
    virtual functor_caller_base<R,T1,void>* clone() const { return new functor_caller<F,R,T1,void>(f); } 
private:
    F f;
};

template< typename R, typename T1 >
struct function< R, T1, void >
{
    function() : fc(0) {}
    ~function() { delete fc; }
    template< typename F > function( F f ) { fc = new functor_caller<F,R,T1,void>(f); }
    function( const function& r ) : fc(0) { operator=(r); }
    function& operator=( const function& r )
    {
        if( this == &r ) return *this;
        delete fc;
        fc = r.fc->clone();
        return *this;
    }
    R operator()( T1 p1 ) const
    {
        smart_assert(fc).msg( "Uninitialized function object" ).fatal();
        return (*fc)( p1 );
    }
private:
    functor_caller_base<R,T1,void> *fc;
};


//////////////////////////////////////////////////////////////////////////////
// no parameters

template< typename R  >
struct functor_caller_base< R, void, void >
{
    virtual ~functor_caller_base() {}
    virtual R operator()() const = 0;
    virtual functor_caller_base* clone() const = 0; 
};

template< typename F, typename R >
struct functor_caller<F,R,void,void> : functor_caller_base<R,void,void>
{
    functor_caller( F f ) : f(f) {}
    virtual R operator()() const { return f(); }
    virtual functor_caller_base<R,void,void>* clone() const { return new functor_caller<F,R,void,void>(f); } 
private:
    F f;
};

template< typename R >
struct function< R, void, void >
{
    function() : fc(0) {}
    ~function() { delete fc; }
    template< typename F > function( F f ) { fc = new functor_caller<F,R,void,void>(f); }
    function( const function& r ) : fc(0) { operator=(r); }
    function& operator=( const function& r )
    {
        if( this == &r ) return *this;
        delete fc;
        fc = r.fc->clone();
        return *this;
    }
    R operator()() const
    {
        smart_assert(fc).msg( "Uninitialized function object" ).fatal();
        return (*fc)();
    }
private:
    functor_caller_base<R,void,void> *fc;
};


#endif _FUNCTION_H_
